#include "daisy_seed.h"
#include "daisysp.h"
#include <memory>
#include "daisy_pod.h"
#include "pedal_board.hpp"

using namespace daisysp;
using namespace daisy;

// CoreSystem クラスの前方宣言
class CoreSystem;

// CoreSystem クラスのインスタンスを保持するユニークポインタ
std::unique_ptr<CoreSystem> core;

// オーディオコールバック関数の前方宣言
void audioCallback(AudioHandle::InputBuffer in, AudioHandle::OutputBuffer out, size_t size);

// CoreSystem クラスの定義
class CoreSystem
{
public:
    CoreSystem() : feedback_(0.5f), delay_time_(0.5f) // ディレイタイム0.5秒（サンプルではなく秒数）
    {
    }

    void init(){
        hw_.Init();
        hw_.SetAudioBlockSize(4); // コールバックごとのサンプル数
        hw_.SetAudioSampleRate(SaiHandle::Config::SampleRate::SAI_48KHZ);
        hw_.StartAdc();
            
        p_knob1.Init(hw_.knob1, 0, 1, Parameter::LINEAR);
        p_knob2.Init(hw_.knob2, 0, 1, Parameter::LINEAR);
        hw_.StartAudio(audioCallback);  
        hw_.seed.StartLog(true);
        hw_.seed.PrintLine("hello world");

        // ディレイエフェクトの初期化
        delay_.Init();
        delay_.SetDelay(delay_time_ * hw_.AudioSampleRate());
        
        knob_1 = p_knob1.Process(); 
        knob_2 = p_knob2.Process(); 
    }

    void updateDelayTime()
    {
        knob_1 = p_knob1.Process(); 
        knob_2 = p_knob2.Process(); 
        // hw_.seed.PrintLine("hello world :");
        // hw_.seed.PrintLine("hello world : %.4f , %.4f",knob_1, knob_2);
        // ディレイタイムとフィードバック量を更新
        delay_time_ = knob_1; // 0～1の範囲をサンプル数にマッピング（最大1秒）
        feedback_ = knob_2;
        delay_.SetDelay(delay_time_ * hw_.AudioSampleRate());
    }

    void run(){
        init();
        while(1) {
            // メインループのコード
            
            // knob_1 = p_knob1.Process(); 
            // knob_2 = p_knob2.Process(); 
            // hw_.seed.PrintLine("hello world : %.4f , %.4f",knob_1, knob_2);
        }
    }

    // インスタンスメンバー関数
    void processAudio(AudioHandle::InputBuffer in, AudioHandle::OutputBuffer out, size_t size)
    {
        hw_.ProcessAllControls();
        // hw_.seed.PrintLine("hello world");
        updateDelayTime();
        float sum = 0.0;
        for (size_t i = 0; i < size; i++)
        {
            applyDelayEffect(in, out, i);
            sum+=in[0][i];
        }
        float total = sum / size;
        hw_.led2.Set(0.0, total, 0.0);
        hw_.led1.Set(knob_1, 0.0, knob_2);
        hw_.UpdateLeds();
    }

private:
    daisy::DaisyPod hw_;
    DelayLine<float, 48000> delay_; // 1秒のディレイバッファ
    float feedback_; // フィードバック量
    float knob_1, knob_2;
    float delay_time_; // ディレイタイム（サンプル数）
    Parameter p_knob1, p_knob2;
    
    float level_;

    // ディレイエフェクトを適用するメンバー関数
    void applyDelayEffect(AudioHandle::InputBuffer in, AudioHandle::OutputBuffer out, size_t index)
    {
        float dry_signal = in[0][index];
        dry_signal = dry_signal + (dry_signal * 100.0 * knob_1);
        if(dry_signal < 0.0){
            dry_signal = std::max(dry_signal, -knob_2);
        }else{
            dry_signal = std::min(dry_signal, knob_2 + 0.1f);
        }
        // float wet_signal = delay_.Read();
        // delay_.Write(dry_signal + (wet_signal * feedback_));
        // delay_.Write(dry_signal);
        out[0][index] =  dry_signal;
        out[1][index] =  dry_signal; // 両方のチャンネルに同じ信号を出力
    }
};

// オーディオコールバック関数の定義
void audioCallback(AudioHandle::InputBuffer in, AudioHandle::OutputBuffer out, size_t size)
{
    if (core)
    {
        core->processAudio(in, out, size);
    }
}
